"""Unit tests for the Lexicon CLI parser"""
# pylint: disable=missing-docstring
from __future__ import absolute_import

import pytest
from lexicon.parser import (
    generate_base_provider_parser,
    generate_cli_main_parser,
)


def test_base_provider_parser():
    baseparser = generate_base_provider_parser()
    parsed = baseparser.parse_args(['list', 'capsulecd.com', 'TXT'])
    assert parsed.action == 'list'
    assert parsed.domain == 'capsulecd.com'
    assert parsed.type == 'TXT'
    assert parsed.ttl is None
    assert parsed.output == 'TABLE'


def test_base_provider_parser_without_domain():
    baseparser = generate_base_provider_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args(['list'])


def test_base_provider_parser_without_options():
    baseparser = generate_base_provider_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args([])


def test_cli_main_parser():
    baseparser = generate_cli_main_parser()
    parsed = baseparser.parse_args(
        ['cloudflare', 'list', 'capsulecd.com', 'TXT'])
    assert parsed.provider_name == 'cloudflare'
    assert parsed.action == 'list'
    assert parsed.domain == 'capsulecd.com'
    assert parsed.type == 'TXT'
    assert parsed.output == 'TABLE'


def test_cli_main_parser_without_args():
    baseparser = generate_cli_main_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args([])



"""Integration tests for Hetzner"""
from unittest import TestCase
import os

import mock
import pytest

# Optional dependency handling
try:
    from bs4 import BeautifulSoup
except ImportError:
    pass

from lexicon.tests.providers.integration_tests import IntegrationTests


def _no_dns_lookup():
    try:
        import dns.resolver  # pylint: disable=import-outside-toplevel
    except ImportError:
        return False

    _domains = ['rimek.info', 'bettilaila.com']
    _resolver = dns.resolver.Resolver()
    _resolver.lifetime = 1
    try:
        for _domain in _domains:
            _ = dns.resolver.zone_for_name(_domain, resolver=_resolver)
        return False
    except dns.exception.DNSException:
        pass
    return True


class HetznerIntegrationTests(IntegrationTests):
    """Base TestCase for Hetzner"""
    @pytest.fixture(autouse=True)
    def _dns_cname_mock(self, request):
        _ignore_mock = request.node.get_closest_marker('ignore_dns_cname_mock')
        _domain_mock = self.domain
        if request.node.name == 'test_provider_authenticate_with_unmanaged_domain_should_fail':
            _domain_mock = 'thisisadomainidonotown.com'
        if _ignore_mock:
            yield
        else:
            with mock.patch('lexicon.providers.hetzner.Provider._get_dns_cname',
                            return_value=(_domain_mock, [], None)) as fixture:
                yield fixture

    @pytest.mark.skipif(_no_dns_lookup(), reason='No DNS resolution possible.')
    @pytest.mark.ignore_dns_cname_mock
    def _test_get_dns_cname(self):
        """Ensure that zone for name can be resolved through dns.resolver call."""
        _domain, _nameservers, _cname = self.provider_module.Provider._get_dns_cname(  # pylint: disable=protected-access
            ('_acme-challenge.fqdn.{}.'.format(self.domain)), False)

        assert _domain == self.domain
        assert _nameservers
        assert not _cname


class HetznerRobotProviderTests(TestCase, HetznerIntegrationTests):
    """TestCase for Hetzner Robot"""
    provider_name = 'hetzner'
    provider_variant = 'Robot'
    domain = 'rimek.info'

    def _filter_post_data_parameters(self):
        return ['_username', '_password', '_csrf_token']

    def _filter_headers(self):
        return ['Cookie']

    def _filter_response(self, response):
        for cookie in ['set-cookie', 'Set-Cookie']:
            if cookie in response['headers']:
                del response['headers'][cookie]
        if os.environ.get('LEXICON_LIVE_TESTS', 'false') == 'true':
            filter_body = (BeautifulSoup(response['body']['string'], 'html.parser')
                           .find(id='center_col'))
            if not filter_body:
                filter_body = (BeautifulSoup(response['body']['string'], 'html.parser')
                               .find(id='login-form'))
            response['body']['string'] = str(filter_body).encode('UTF-8')
        return response

    def _test_parameters_overrides(self):
        options = {'auth_account': 'robot',
                   'linked': 'no',
                   'propagated': 'no',
                   'latency': 0.00001}
        return options


class HetznerKonsoleHProviderTests(TestCase, HetznerIntegrationTests):
    """TestCase for KonsoleH"""
    provider_name = 'hetzner'
    provider_variant = 'KonsoleH'
    domain = 'bettilaila.com'

    def _filter_post_data_parameters(self):
        return ['login_user_inputbox', 'login_pass_inputbox', '_csrf_name', '_csrf_token']

    def _filter_headers(self):
        return ['Cookie']

    def _filter_response(self, response):
        for cookie in ['set-cookie', 'Set-Cookie']:
            if cookie in response['headers']:
                del response['headers'][cookie]
        if os.environ.get('LEXICON_LIVE_TESTS', 'false') == 'true':
            filter_body = (BeautifulSoup(response['body']['string'], 'html.parser')
                           .find(id='content'))
            if not filter_body:
                filter_body = (BeautifulSoup(response['body']['string'], 'html.parser')
                               .find(id='loginform'))
            response['body']['string'] = str(filter_body).encode('UTF-8')
        return response

    def _test_parameters_overrides(self):
        env_username = os.environ.get('LEXICON_HETZNER_KONSOLEH_USERNAME', 'placeholder_username')
        env_password = os.environ.get('LEXICON_HETZNER_KONSOLEH_PASSWORD', 'placeholder_password')
        options = {'auth_account': 'konsoleh',
                   'auth_username': env_username,
                   'auth_password': env_password,
                   'linked': 'no',
                   'propagated': 'no',
                   'latency': 0.00001}
        return options



"""Module provider for Godaddy"""
from __future__ import absolute_import
import hashlib
import json
import logging

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['godaddy.com']


def provider_parser(subparser):
    """Generate a subparser for Godaddy"""
    subparser.add_argument(
        '--auth-key', help='specify the key to access the API')
    subparser.add_argument(
        '--auth-secret', help='specify the secret to access the API')


class Provider(BaseProvider):
    """
    Implements the DNS GoDaddy provider.
    Some general remarks about this provider, because it uses a weirdly designed API.
    Indeed, there is no direct way to insert, update or delete a specific record.
    Furthermore, there is no unique identifier for a record.
    Instead GoDaddy use a replace approach: for a given set of records one
    can replace this set with a new set sent through API.
    For the sake of simplicity and consistency across the provider edit methods,
    the set will be always all records in the DNS zone.
    With this approach:
        - adding a record consists in appending a record to the obtained set and call
          replace with the updated set,
        - updating a record consists in modifying a record in the obtained set and call
          replace with the updated set,
        - deleting a record consists in removing a record in the obtained set and call
          replace with the updated set.
    In parallel, as said before, there is no unique identifier.
    This provider then implement a pseudo-identifier, to allow an easy update or delete
    using the '--identifier' lexicon parameter.
    But you need to call the 'list' command just before executing and update/delete action,
    because identifier value is tied to the content of the record, and will change anytime
    something is changed in the record.
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.godaddy.com/v1'

    def _authenticate(self):
        domain = self.domain

        result = self._get('/domains/{0}'.format(domain))
        self.domain_id = result['domainId']

    def _list_records(self, rtype=None, name=None, content=None):
        domain = self.domain

        url = '/domains/{0}/records'.format(domain)
        if rtype:
            url += '/{0}'.format(rtype)
        if name:
            url += '/{0}'.format(self._relative_name(name))

        raws = self._get(url)

        records = []
        for raw in raws:
            records.append({
                'id': Provider._identifier(raw),
                'type': raw['type'],
                'name': self._full_name(raw['name']),
                'ttl': raw['ttl'],
                'content': raw['data']
            })

        if content:
            records = [
                record for record in records if record['data'] == content]

        LOGGER.debug('list_records: %s', records)

        return records

    def _create_record(self, rtype, name, content):
        domain = self.domain
        relative_name = self._relative_name(name)
        ttl = self._get_lexicon_option('ttl')

        # Retrieve existing data in DNS zone.
        records = self._get('/domains/{0}/records'.format(domain))

        # Check if a record already matches given parameters
        for record in records:
            if (record['type'] == rtype and self._relative_name(record['name']) == relative_name
                    and record['data'] == content):
                LOGGER.debug(
                    'create_record (ignored, duplicate): %s %s %s', rtype, name, content)
                return True

        # Append a new entry corresponding to given parameters.
        data = {'type': rtype, 'name': relative_name, 'data': content}
        if ttl:
            data['ttl'] = ttl

        records.append(data)

        # Synchronize data with inserted record into DNS zone.
        self._put('/domains/{0}/records'.format(domain), records)

        LOGGER.debug('create_record: %s %s %s', rtype, name, content)

        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # No identifier is used with GoDaddy.
        # We can rely either:
        #   - only on rtype/name to get the relevant records, both of them are required
        #     or we will could update to much records ...,
        #   - or by the pseudo-identifier provided
        # Furthermore for rtype/name approach, we cannot update all matching records, as it
        # would lead o an error (two entries of same rtype + name cannot have the same content).
        # So for rtype/name approach, we search first matching record for rtype/name on which
        # content is different, and we update it before synchronizing the DNS zone.
        if not identifier and not rtype:
            raise Exception('ERROR: rtype is required')
        if not identifier and not name:
            raise Exception('ERROR: name is required')

        domain = self.domain
        relative_name = None
        if name:
            relative_name = self._relative_name(name)

        # Retrieve existing data in DNS zone.
        records = self._get('/domains/{0}/records'.format(domain))

        # Get the record to update:
        #   - either explicitly by its identifier,
        #   - or the first matching by its rtype+name where content does not match
        #     (first match, see first method comment for explanation).
        for record in records:
            if ((identifier and Provider._identifier(record) == identifier) or  # pylint: disable=too-many-boolean-expressions
                    (not identifier and record['type'] == rtype
                     and self._relative_name(record['name']) == relative_name
                     and record['data'] != content)):
                record['data'] = content
                break

        # Synchronize data with updated records into DNS zone.
        self._put('/domains/{0}/records'.format(domain), records)

        LOGGER.debug('update_record: %s %s %s', rtype, name, content)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # For the LOL. GoDaddy does not accept an empty array
        # when updating a particular set of records.
        # It means that you cannot request to remove all records
        # matching a particular rtype and/or name.
        # Instead, we get ALL records in the DNS zone, update the set,
        # and replace EVERYTHING in the DNS zone.
        # You will always have at minimal NS/SRV entries in the array,
        # otherwise your DNS zone is broken, and updating the zone is the least of your problem ...
        domain = self.domain

        # Retrieve all records in the DNS zone
        records = self._get('/domains/{0}/records'.format(domain))

        relative_name = None
        if name:
            relative_name = self._relative_name(name)

        # Filter out all records which matches the pattern (either identifier
        # or some combination of rtype/name/content).
        filtered_records = []
        if identifier:
            filtered_records = [
                record for record in records if Provider._identifier(record) != identifier]
        else:
            for record in records:
                if ((not rtype and not relative_name and not content)  # pylint: disable=too-many-boolean-expressions
                        or (rtype and not relative_name and not content and record['type'] != rtype)
                        or (not rtype and relative_name and not content
                            and self._relative_name(record['name']) != relative_name)
                        or (not rtype and not relative_name and content
                            and record['data'] != content)
                        or (rtype and relative_name and not content
                            and (record['type'] != rtype
                                 or self._relative_name(record['name']) != relative_name))
                        or (rtype and not relative_name and content
                            and (record['type'] != rtype or record['data'] != content))
                        or (not rtype and relative_name and content
                            and (self._relative_name(record['name']) != relative_name
                                 or record['data'] != content))
                        or (rtype and relative_name and content
                            and (record['type'] != rtype
                                 or self._relative_name(record['name']) != relative_name
                                 or record['data'] != content))):
                    filtered_records.append(record)

        # Synchronize data with expurged entries into DNS zone.
        self._put('/domains/{0}/records'.format(domain), filtered_records)

        LOGGER.debug('delete_records: %s %s %s', rtype, name, content)

        return True

    # GoDaddy provides no identifier for a record, which is a problem
    # where identifiers can be used (delete and update).
    # To circumvent this, we implement a pseudo-identifier,which is basically
    # a hash of type+name+content of a given record.
    # It is far from perfect, as the identifier will change each time
    # we change something in the record ...
    # But at least, one can use 'lexicon godaddy list ...' then
    # 'lexicon godaddy update --identifier ...' to modify specific record.
    # However, 'lexicon godaddy list ...' should be called each time DNS
    # zone had been changed to calculate new identifiers.
    @staticmethod
    def _identifier(record):
        sha256 = hashlib.sha256()
        sha256.update(('type=' + record.get('type', '') + ',').encode('utf-8'))
        sha256.update(('name=' + record.get('name', '') + ',').encode('utf-8'))
        sha256.update(('data=' + record.get('data', '') + ',').encode('utf-8'))
        return sha256.hexdigest()[0:7]

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if not data:
            data = {}
        if not query_params:
            query_params = {}

        # When editing DNS zone, API is unavailable for few seconds
        # (until modifications are propagated).
        # In this case, call to API will return 409 HTTP error.
        # We use the Retry extension to retry the requests until
        # we get a processable reponse (402 HTTP status, or an HTTP error != 409)
        retries = Retry(
            total=10,
            backoff_factor=0.5,
            status_forcelist=[409],
            method_whitelist=frozenset(
                ['GET', 'PUT', 'POST', 'DELETE', 'PATCH'])
        )

        session = requests.Session()
        session.mount('https://', HTTPAdapter(max_retries=retries))

        result = session.request(action, self.api_endpoint + url,
                                 params=query_params,
                                 data=json.dumps(data),
                                 headers={
                                     'Content-Type': 'application/json',
                                     'Accept': 'application/json',
                                     # GoDaddy use a key/secret pair to authenticate
                                     'Authorization': 'sso-key {0}:{1}'.format(
                                         self._get_provider_option(
                                             'auth_key'),
                                         self._get_provider_option('auth_secret'))
                                 })

        result.raise_for_status()

        try:
            # Return the JSON body response if exists.
            return result.json()
        except ValueError:
            # For some requests command (eg. PUT), GoDaddy will not
            # return any JSON, just an HTTP status without body.
            return None



"""Module provider for OVH"""
from __future__ import absolute_import
import hashlib
import json
import logging
import time

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

ENDPOINTS = {
    'ovh-eu': 'https://eu.api.ovh.com/1.0',
    'ovh-ca': 'https://ca.api.ovh.com/1.0',
    'ovh-us': 'https://api.ovhcloud.com/1.0',
    'kimsufi-eu': 'https://eu.api.kimsufi.com/1.0',
    'kimsufi-ca': 'https://ca.api.kimsufi.com/1.0',
    'soyoustart-eu': 'https://eu.api.soyoustart.com/1.0',
    'soyoustart-ca': 'https://ca.api.soyoustart.com/1.0',
}

NAMESERVER_DOMAINS = ['ovh.net', 'anycast.me']


def provider_parser(subparser):
    """Generate provider parser for OVH"""
    subparser.description = '''
        OVH Provider requires a token with full rights on /domain/*.
        It can be generated for your OVH account on the following URL:
        https://api.ovh.com/createToken/index.cgi?GET=/domain/*&PUT=/domain/*&POST=/domain/*&DELETE=/domain/*'''
    subparser.add_argument('--auth-entrypoint', help='specify the OVH entrypoint', choices=[
        'ovh-eu', 'ovh-ca', 'soyoustart-eu', 'soyoustart-ca', 'kimsufi-eu', 'kimsufi-ca'
    ])
    subparser.add_argument('--auth-application-key',
                           help='specify the application key')
    subparser.add_argument('--auth-application-secret',
                           help='specify the application secret')
    subparser.add_argument('--auth-consumer-key',
                           help='specify the consumer key')


class Provider(BaseProvider):
    """Provider class for OVH"""
    def __init__(self, config):
        super(Provider, self).__init__(config)

        # Handling missing required parameters
        if not self._get_provider_option('auth_entrypoint'):
            raise Exception('Error, entrypoint is not defined')
        if not self._get_provider_option('auth_application_key'):
            raise Exception('Error, application key is not defined')
        if not self._get_provider_option('auth_application_secret'):
            raise Exception('Error, application secret is not defined')
        if not self._get_provider_option('auth_consumer_key'):
            raise Exception('Error, consumer key is not defined')

        # Construct DNS OVH environment
        self.domain_id = None
        self.endpoint_api = ENDPOINTS.get(
            self._get_provider_option('auth_entrypoint'))
        self.session = None
        self.time_delta = None

    def _authenticate(self):
        # All requests will be done in one HTTPS session
        self.session = requests.Session()

        # Calculate delta time between local and OVH to avoid requests rejection
        server_time = self.session.get(
            '{0}/auth/time'.format(self.endpoint_api)).json()
        self.time_delta = server_time - int(time.time())

        # Get domain and status
        domain = self.domain

        domains = self._get('/domain/zone/')
        if domain not in domains:
            raise Exception('Domain {0} not found'.format(domain))

        status = self._get('/domain/zone/{0}/status'.format(domain))
        if not status['isDeployed']:
            raise Exception('Zone {0} is not deployed'.format(domain))

        self.domain_id = domain

    def _create_record(self, rtype, name, content):
        domain = self.domain
        ttl = self._get_lexicon_option('ttl')

        records = self._list_records(rtype, name, content)
        for record in records:
            if (record['type'] == rtype
                    and self._relative_name(record['name']) == self._relative_name(name)
                    and record['content'] == content):
                LOGGER.debug(
                    'create_record (ignored, duplicate): %s %s %s', rtype, name, content)
                return True

        data = {
            'fieldType': rtype,
            'subDomain': self._relative_name(name),
            'target': content
        }

        if ttl:
            data['ttl'] = ttl

        result = self._post('/domain/zone/{0}/record'.format(domain), data)
        self._post('/domain/zone/{0}/refresh'.format(domain))

        LOGGER.debug('create_record: %s', result['id'])

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        domain = self.domain
        records = []

        params = {}
        if rtype:
            params['fieldType'] = rtype
        if name:
            params['subDomain'] = self._relative_name(name)

        record_ids = self._get(
            '/domain/zone/{0}/record'.format(domain), params)

        for record_id in record_ids:
            raw = self._get(
                '/domain/zone/{0}/record/{1}'.format(domain, record_id))
            records.append({
                'type': raw['fieldType'],
                'name': self._full_name(raw['subDomain']),
                'ttl': raw['ttl'],
                'content': raw['target'],
                'id': raw['id']
            })

        if content:
            records = [
                record for record in records if record['content'].lower() == content.lower()]

        LOGGER.debug('list_records: %s', records)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        domain = self.domain

        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]['id']
            elif len(records) > 1:
                raise Exception('Several record identifiers match the request')
            else:
                raise Exception('Record identifier could not be found')

        data = {}
        if name:
            data['subDomain'] = self._relative_name(name)
        if content:
            data['target'] = content

        self._put(
            '/domain/zone/{0}/record/{1}'.format(domain, identifier), data)
        self._post('/domain/zone/{0}/refresh'.format(domain))

        LOGGER.debug('update_record: %s', identifier)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        domain = self.domain

        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self._delete(
                '/domain/zone/{0}/record/{1}'.format(domain, record_id))

        self._post('/domain/zone/{0}/refresh'.format(domain))

        LOGGER.debug('delete_record: %s', True)

        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        headers = {}
        target = self.endpoint_api + url
        body = ''

        if data is not None:
            headers['Content-type'] = 'application/json'
            body = json.dumps(data)

        # Get correctly sync time
        now = str(int(time.time()) + self.time_delta)

        headers['X-Ovh-Application'] = self._get_provider_option(
            'auth_application_key')
        headers['X-Ovh-Consumer'] = self._get_provider_option(
            'auth_consumer_key')
        headers['X-Ovh-Timestamp'] = now

        request = requests.Request(
            action, target, data=body, params=query_params, headers=headers)
        prepared_request = self.session.prepare_request(request)

        # Build OVH API signature for the current request
        signature = hashlib.sha1()
        signature.update('+'.join([
            self._get_provider_option('auth_application_secret'),
            self._get_provider_option('auth_consumer_key'),
            action.upper(),
            prepared_request.url,
            body,
            now
        ]).encode('utf-8'))

        # Sign the request
        headers['X-Ovh-Signature'] = '$1$' + signature.hexdigest()

        result = self.session.request(
            method=action,
            url=target,
            params=query_params,
            data=body,
            headers=headers
        )
        result.raise_for_status()

        return result.json()



"""Module provider for Dreamhost"""
from __future__ import absolute_import
import base64
import json
import logging
import time

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['dreamhost.com']

_DATA_NON_EXIST_ERROR_LIST = [
    'no_record',
    'no_type',
    'no_value',
    'no_such_record',
    'no_such_type',
    'no_such_value',
    'no_such_zone',
]

_DATA_ALREADY_EXIST_ERROR_LIST = [
    'record_already_exists_not_editable',
    'record_already_exists_remove_first',
    'CNAME_already_on_record',
]


class NonExistError(Exception):
    """NonExistError"""


class AlreadyExistError(Exception):
    """AlreadyExistError"""


def provider_parser(subparser):
    """Module provider for Dreamhost"""
    subparser.add_argument(
        "--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Dreamhost"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.dreamhost.com/'

    # Dreamhost provides no identifier for a record.
    # Furthermore, Dreamhost requires type, record, value to delete a record.
    # The record defined in lexicon is {type, name, content, id}
    # We use base64(json({'type', 'name', 'content'}))
    # as the identifier of Dreamhost record.
    @staticmethod
    def _identifier(dreamhost_record):
        id_struct = {
            'type': dreamhost_record['type'],
            'name': dreamhost_record['record'],
            'content': dreamhost_record['value'],
        }
        return base64.urlsafe_b64encode(json.dumps(id_struct).encode('utf-8')).decode('utf-8')

    # The information in identifier follows the record in lexicon.
    # Provider._record_to_dreamhost_record transfers to dreamhost-based record.
    @staticmethod
    def _id_to_dreamhost_record(identifier):
        record = json.loads(base64.urlsafe_b64decode(identifier.encode('utf-8')).decode('utf-8'))
        dreamhost_record = Provider._record_to_dreamhost_record(record)
        return dreamhost_record

    # The information in identifier follows the record in lexicon.
    # 'id' is added in the record.
    @staticmethod
    def _id_to_record(identifier):
        record = json.loads(base64.urlsafe_b64decode(identifier.encode('utf-8')).decode('utf-8'))
        record['id'] = identifier

        return record

    # Transferring lexicon-based record to Dreamhost-based record.
    @staticmethod
    def _record_to_dreamhost_record(record):
        dreamhost_record = {
            'type': record['type'],
            'record': record['name'],
            'value': record['content'],
        }
        return dreamhost_record

    def _authenticate(self):
        self.domain_id = None
        payload = self._get('domain-list_domains')
        data = payload.get('data', None)
        if data is None:
            raise Exception('Domain not found')

        for domain in data:
            if domain.get('domain', '') == self.domain:
                self.domain_id = self.domain
        if self.domain_id is None:
            raise Exception('Domain not found')

    def _create_record(self, rtype, name, content):
        name = self._full_name(name)

        try:
            self._get('dns-add_record', query_params={
                'record': name,
                'type': rtype,
                'value': content,
            })
        except AlreadyExistError:
            pass

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):

        payload = self._get('dns-list_records')

        resource_list = payload.get('data', None)
        if not isinstance(resource_list, list):
            raise Exception('unable to get records: %s' % payload)

        resource_list = [
            resource for resource in resource_list if resource['zone'] == self.domain]
        if rtype:
            resource_list = [
                resource for resource in resource_list if resource['type'] == rtype]
        if name:
            name = self._full_name(name)
            resource_list = [
                resource for resource in resource_list if resource['record'] == name]
        if content:
            resource_list = [
                resource for resource in resource_list if resource['value'] == content]

        processed_records = []
        for dreamhost_record in resource_list:
            processed_records.append({
                'id': Provider._identifier(dreamhost_record),
                'type': dreamhost_record['type'],
                'name': dreamhost_record['record'],
                'content': dreamhost_record['value'],
            })

        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            try:
                self._delete_record(identifier)
            except NonExistError:
                pass

        return self._create_record(rtype=rtype, name=name, content=content)

    # Delete existing records.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        to_deletes = []
        if identifier:
            record = Provider._id_to_record(identifier)
            to_deletes.append(record)
        else:
            records = self._list_records(rtype=rtype, name=name, content=content)
            to_deletes = records

        # for-loop to delete deletes.
        err = None
        for each in to_deletes:
            try:
                dreamhost_record = Provider._record_to_dreamhost_record(each)
                self._get('dns-remove_record', query_params=dreamhost_record)

            except Exception as exception:  # pylint: disable=broad-except
                err = exception

            # Sleeping for 1-second to avoid trigerring ddos protecting in case of looped requests
            time.sleep(1)

        if err is not None:
            raise err

        return True

    # Helpers
    def _request(self, action='GET', url='', data=None, query_params=None):
        if data is None:
            data = {}

        if query_params is None:
            query_params = {}

        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

        query_params['key'] = self._get_provider_option('auth_token')
        query_params['format'] = 'json'
        if 'cmd' not in query_params:
            query_params['cmd'] = url

        response = requests.request(action, self.api_endpoint, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers)

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        result = response.json()
        if result.get('result', '') != 'success':
            err_msg = result.get('data', '')
            if err_msg in _DATA_NON_EXIST_ERROR_LIST:
                raise NonExistError('Dreamhost non-exist error: {0}'.format(result))
            if err_msg in _DATA_ALREADY_EXIST_ERROR_LIST:
                raise AlreadyExistError('Dreamhost already-exist error: {0}'.format(result))
            raise Exception('Dreamhost api error: {0}'.format(result))
        return result



"""Integration tests for PowerDNS"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class PowerdnsProviderTests(TestCase, IntegrationTests):
    """TestCase for PowerDNS"""
    provider_name = 'powerdns'
    domain = 'sometestdomain.com'

    def _filter_headers(self):
        return ['X-API-Key']

    def _test_parameters_overrides(self):
        return {'pdns_server': 'http://127.0.0.1:8081', 'pdns_server_id': 'localhost'}

    # TODO: the following skipped suite and fixtures should be enabled
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return



"""Module provider for DNSPod"""
from __future__ import absolute_import

import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['dnsapi.cn']


def provider_parser(subparser):
    """Configure provider parser for DNSPod"""
    subparser.add_argument(
        "--auth-username", help="specify api id for authentication")
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for DNSPod"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://dnsapi.cn'

    def _authenticate(self):

        payload = self._post('/Domain.Info', {'domain': self.domain})

        if payload['status']['code'] != '1':
            raise Exception(payload['status']['message'])

        self.domain_id = payload['domain']['id']

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        record = {
            'domain_id': self.domain_id,
            'sub_domain': self._relative_name(name),
            'record_type': rtype,
            'record_line': u'\u9ED8\u8BA4',
            'value': content
        }
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')

        payload = self._post('/Record.Create', record)

        if payload['status']['code'] not in ['1', '31']:
            raise Exception(payload['status']['message'])

        LOGGER.debug('create_record: %s', payload['status']['code'] == '1')
        return payload['status']['code'] == '1'

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._post('/Record.List', {'domain': self.domain})
        LOGGER.debug('payload: %s', payload)
        records = []
        for record in payload['records']:
            processed_record = {
                'type': record['type'],
                'name': self._full_name(record['name']),
                'ttl': record['ttl'],
                'content': record['value'],
                # this id is useless unless your doing record linking.
                # Lets return the original record identifier.
                'id': record['id']
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record['type'] == rtype]
        if name:
            records = [record for record in records if record['name']
                       == self._full_name(name)]
        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {
            'domain_id': self.domain_id,
            'record_id': identifier,
            'sub_domain': self._relative_name(name),
            'record_type': rtype,
            'record_line': u'\u9ED8\u8BA4',
            'value': content
        }
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        LOGGER.debug('data: %s', data)
        payload = self._post('/Record.Modify', data)
        LOGGER.debug('payload: %s', payload)
        if payload['status']['code'] != '1':
            raise Exception(payload['status']['message'])

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self._post(
                '/Record.Remove', {'domain_id': self.domain_id, 'record_id': record_id})

            # if payload['status']['code'] != '1':
            #    raise Exception(payload['status']['message'])

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        data['login_token'] = self._get_provider_option(
            'auth_username') + ',' + self._get_provider_option('auth_token')
        data['format'] = 'json'
        if query_params is None:
            query_params = {}
        default_headers = {}
        default_auth = None
        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=data,
                                    headers=default_headers,
                                    auth=default_auth)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Sakura Cloud"""
from __future__ import absolute_import
import json
import logging

import requests
from requests.auth import HTTPBasicAuth

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['sakura.ne.jp']


def provider_parser(subparser):
    """Generate a provider parser for Sakura Cloud"""
    subparser.add_argument(
        "--auth-token", help="specify access token for authentication")
    subparser.add_argument(
        "--auth-secret", help="specify access secret for authentication")


class Provider(BaseProvider):
    """Provider class for Sakura Cloud"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://secure.sakura.ad.jp/cloud/zone/is1a/api/cloud/1.1'

    def _authenticate(self):

        query_params = {
            "Filter": {
                "Provider.Class": "dns",
                "Name": self.domain
            }
        }
        payload = self._get('/commonserviceitem', query_params=query_params)

        for item in payload["CommonServiceItems"]:
            if item["Status"]["Zone"] == self.domain:
                self.domain_id = item["ID"]
                return

        raise Exception('No domain found')

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        name = self._relative_name(name)
        resource_record_sets = self._get_resource_record_sets()
        index = self._find_resource_record_set(
            resource_record_sets, rtype=rtype, name=name, content=content)
        if index >= 0:
            LOGGER.debug('create_record: %s', False)
            return False

        resource_record_sets.append(
            {
                "Name": name,
                "Type": rtype,
                "RData": self._bind_format_target(rtype, content),
                "TTL": self._get_lexicon_option('ttl'),
            }
        )

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        records = []

        for record in self._get_resource_record_sets():
            processed_record = {
                'type': record['Type'],
                'name': self._full_name(record['Name']),
                'ttl': record['TTL'],
                'content': record['RData'],
                # 'id': None,
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record['type'] == rtype]
        if name:
            records = [
                record for record in records if record['name'] == self._full_name(name)
            ]
        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):

        if not (rtype and name and content):
            raise Exception("rtype ,name and content must be specified.")

        name = self._relative_name(name)
        resource_record_sets = self._get_resource_record_sets()
        index = self._find_resource_record_set(
            resource_record_sets, rtype=rtype, name=name)

        if index >= 0:
            resource_record_sets[index]["Type"] = rtype
            resource_record_sets[index]["Name"] = name
            resource_record_sets[index]["RData"] = self._bind_format_target(
                rtype, content)
            resource_record_sets[index]["TTL"] = self._get_lexicon_option(
                'ttl')
        else:
            resource_record_sets.append(
                {
                    "Name": name,
                    "Type": rtype,
                    "RData": self._bind_format_target(rtype, content),
                    "TTL": self._get_lexicon_option('ttl'),
                }
            )

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug('create_record')

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        resource_record_sets = self._get_resource_record_sets()

        if name is not None:
            name = self._relative_name(name)
        if content is not None:
            content = self._bind_format_target(rtype, content)

        filtered_records = []
        for record in resource_record_sets:
            if rtype and record['Type'] != rtype:
                continue
            if name and record['Name'] != name:
                continue
            if content and record['RData'] != content:
                continue
            filtered_records.append(record)

        if not filtered_records:
            LOGGER.debug('delete_record: %s', False)
            return False

        for record in filtered_records:
            resource_record_sets.remove(record)

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers
    def _full_name(self, record_name):
        if record_name == "@":
            record_name = self.domain
        return super(Provider, self)._full_name(record_name)

    def _relative_name(self, record_name):
        name = super(Provider, self)._relative_name(record_name)
        if not name:
            name = "@"
        return name

    def _bind_format_target(self, rtype, target):  # pylint: disable=no-self-use
        if rtype == "CNAME" and not target.endswith("."):
            target += "."
        return target

    def _find_resource_record_set(self, records, rtype=None, name=None, content=None):  # pylint: disable=no-self-use
        for index, record in enumerate(records):
            if rtype and record['Type'] != rtype:
                continue
            if name and record['Name'] != name:
                continue
            if content and record['RData'] != content:
                continue
            return index
        return -1

    def _get_resource_record_sets(self):
        payload = self._get('/commonserviceitem/{0}'.format(self.domain_id))
        return payload['CommonServiceItem']['Settings']['DNS']['ResourceRecordSets']

    def _update_resource_record_sets(self, resource_record_sets):
        content = {
            "CommonServiceItem": {
                "Settings": {
                    "DNS": {
                        "ResourceRecordSets": resource_record_sets
                    }
                }
            }
        }
        return self._put('/commonserviceitem/{0}'.format(self.domain_id), content)

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }
        default_auth = HTTPBasicAuth(
            self._get_provider_option('auth_token'), self._get_provider_option('auth_secret'))

        query_string = ""
        if query_params:
            query_string = json.dumps(query_params)

        response = requests.request(action, self.api_endpoint + url, params=query_string,
                                    data=json.dumps(data),
                                    headers=default_headers,
                                    auth=default_auth)
        try:
            # if the request fails for any reason, throw an error.
            response.raise_for_status()
        except BaseException:
            LOGGER.error(response.json().get("error_msg"))
            raise
        return response.json()



#!/usr/bin/env python
"""Module for Lexicon command-line interface"""
from __future__ import absolute_import, print_function
import json
import logging
import os
import sys

from lexicon.client import Client
from lexicon.config import ConfigResolver
from lexicon.parser import generate_cli_main_parser


logger = logging.getLogger(__name__)  # pylint: disable=C0103


def generate_list_table_result(lexicon_logger, output=None, without_header=None):
    """Convert returned data from list actions into a nice table for command line usage"""
    if not isinstance(output, list):
        lexicon_logger.debug('Command output is not a list, and then cannot '
                             'be printed with --quiet parameter not enabled.')
        return None

    array = [[
        row.get('id', ''),
        row.get('type', ''),
        row.get('name', ''),
        row.get('content', ''),
        row.get('ttl', '')] for row in output]

    # Insert header (insert before calculating the max width of each column
    # to take headers size into account)
    if not without_header:
        headers = ['ID', 'TYPE', 'NAME', 'CONTENT', 'TTL']
        array.insert(0, headers)

    column_widths = [0, 0, 0, 0, 0]
    # Find max width for each column
    for row in array:
        for idx, col in enumerate(row):
            width = len(str(col))
            if width > column_widths[idx]:
                column_widths[idx] = width

    # Add a 'nice' separator
    if not without_header:
        array.insert(1, ['-' * column_widths[idx]
                         for idx in range(len(column_widths))])

    # Construct table to be printed
    table = []
    for row in array:
        row_list = []
        for idx, col in enumerate(row):
            row_list.append(str(col).ljust(column_widths[idx]))
        table.append(' '.join(row_list))

    # Return table
    return os.linesep.join(table)


def generate_table_results(output=None, without_header=None):
    """Convert returned data from non-list actions into a nice table for command line usage"""
    array = []
    str_output = str(output)

    if not without_header:
        array.append('RESULT')
        array.append('-' * max(6, len(str_output)))

    array.append(str_output)
    return os.linesep.join(array)


def handle_output(results, output_type, action):
    """Print the relevant output for given output_type"""
    if output_type == 'QUIET':
        return

    if not output_type == 'JSON':
        if action == 'list':
            table = generate_list_table_result(
                logger, results, output_type == 'TABLE-NO-HEADER')
        else:
            table = generate_table_results(results, output_type == 'TABLE-NO-HEADER')
        if table:
            print(table)
    else:
        try:
            json_str = json.dumps(results)
            if json_str:
                print(json_str)
        except TypeError:
            logger.debug('Output is not JSON serializable, and then cannot '
                         'be printed with --output=JSON parameter.')


def main():
    """Main function of Lexicon."""
    # Dynamically determine all the providers available and gather command line arguments.
    parsed_args = generate_cli_main_parser().parse_args()

    log_level = logging.getLevelName(parsed_args.log_level)
    logging.basicConfig(stream=sys.stdout, level=log_level,
                        format='%(message)s')
    logger.debug('Arguments: %s', parsed_args)

    # In the CLI context, will get configuration interactively:
    #   * from the command line
    #   * from the environment variables
    #   * from lexicon configuration files found in given --config-dir (default is current dir)
    config = ConfigResolver()
    config.with_args(parsed_args).with_env().with_config_dir(parsed_args.config_dir)

    client = Client(config)

    results = client.execute()

    handle_output(results, parsed_args.output, config.resolve('lexicon:action'))


if __name__ == '__main__':
    main()



"""Module provider for NSOne"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['nsone.net']


def provider_parser(subparser):
    """Configure provider parser for NSOne"""
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for NSOne"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.nsone.net/v1'

    def _authenticate(self):

        payload = self._get('/zones/{0}'.format(self.domain))

        if not payload['id']:
            raise Exception('No domain found')

        self.domain_id = self.domain

    def _get_record_set(self, name, rtype):
        try:
            payload = self._get(
                '/zones/{0}/{1}/{2}'.format(self.domain_id, name, rtype))
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            raise

        return {
            'type': payload['type'],
            'name': payload['domain'],
            'ttl': payload['ttl'],
            'answers': payload['answers']
        }

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        name = self._full_name(name)
        existing_record_set = self._get_record_set(name, rtype)
        if existing_record_set:
            def _record_set_has_answer(record_set, content):
                for answer in record_set['answers']:
                    if content in answer['answer']:
                        return True
                return False

            if not _record_set_has_answer(existing_record_set, content):
                existing_record_set['answers'].append({
                    'answer': [content]
                })
                self._post(
                    '/zones/{0}/{1}/{2}'.format(self.domain_id, name, rtype), existing_record_set)
        else:
            record = {
                'type': rtype,
                'domain': name,
                'ttl': self._get_lexicon_option('ttl'),
                'zone': self.domain_id,
                'answers': [
                    {"answer": [content]}
                ]
            }
            payload = {}
            try:
                payload = self._put(
                    '/zones/{0}/{1}/{2}'.format(self.domain_id, name, rtype), record)
            except requests.exceptions.HTTPError as error:
                # http 400 is ok here, because the record probably already exists
                if error.response.status_code != 400:
                    raise

            LOGGER.debug('create_record: %s', 'id' in payload)

        return True

    def _find_record(self, domain, _type=None):
        """search for a record on NS1 across zones. returns None if not found."""

        def _is_matching(record):
            """filter function for records"""

            if domain and record.get('domain', None) != domain:
                return False
            if _type and record.get('type', None) != _type:
                return False
            return True

        payload = self._get('/search?q={0}&type=record'.format(domain))
        for record in payload:
            if _is_matching(record):
                match = record
                break
        else:
            # no such domain on ns1
            return None

        record = self._get(
            '/zones/{0}/{1}/{2}'.format(match['zone'], match['domain'], match['type']))
        if record.get('message', None):
            return None  # {"message":"record not found"}
        short_answers = [x['answer'][0] for x in record['answers']]

        # ensure a compatibility level with self._list_records
        record['short_answers'] = short_answers
        return record

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):

        def _resolve_link(record, recurse=0):
            # https://ns1.com/articles/cname-alias-and-linked-records
            # - recursion is allowed
            # - link source and link target are always of the same rtype
            # - target can be anywhere on ns1, not necessarily self.domain_id.
            if record.get('link', None) is None:
                # not a linked record
                return record

            if recurse < 1:
                return None

            match = self._find_record(record['link'], _type=record['type'])
            if not match:
                return None

            return _resolve_link(match, recurse=recurse - 1)

        payload = self._get('/zones/{0}'.format(self.domain_id))
        records = []
        for record in payload['records']:

            if rtype and record['type'] != rtype:
                continue

            if name and record['domain'] != self._full_name(name):
                continue

            link_target = _resolve_link(record, recurse=3)

            if link_target and link_target.get('short_answers', None):
                # target found (could be the same as orig record)
                answers = link_target['short_answers']
            else:
                # recursion limit reached. or unhandled record format.
                answers = []

            if content and content not in answers:
                continue

            for answer in answers:
                processed_record = {
                    'type': record['type'],
                    'name': record['domain'],
                    'ttl': record['ttl'],
                    'content': answer,
                    # this id is useless unless your doing record linking. Lets return the
                    # original record identifier.
                    'id': '{0}/{1}/{2}'.format(self.domain_id, record['domain'], record['type'])
                }
                records.append(processed_record)

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        new_identifier = "{0}/{1}/{2}".format(
            self.domain_id, self._full_name(name), rtype)

        if(new_identifier == identifier or (rtype is None and name is None)):
            # the identifier hasnt changed, or type and name are both unspecified,
            # only update the content.
            data['answers'] = [
                {"answer": [content]}
            ]
            self._post('/zones/{0}'.format(identifier), data)

        else:
            # identifiers are different
            # get the old record, create a new one with updated data, delete the old record.
            old_record = self._get('/zones/{0}'.format(identifier))
            self.create_record(
                rtype or old_record['type'],
                name or old_record['domain'],
                content or old_record['answers'][0]['answer'][0])
            self.delete_record(identifier)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            name = self._full_name(name)

            record_set = self._get_record_set(name, rtype)
            if record_set:
                record_set_new = {
                    'type': record_set['type'],
                    'name': record_set['name'],
                    'ttl': record_set['ttl'],
                    'answers': []
                }

                if content:
                    for answer in record_set['answers']:
                        if content not in answer['answer']:
                            record_set_new['answers'].append(answer)

                if record_set_new['answers']:
                    self._post(
                        '/zones/{0}/{1}/{2}'.format(self.domain_id, name, rtype), record_set_new)
                else:
                    self._delete(
                        '/zones/{0}/{1}/{2}'.format(self.domain_id, name, rtype))
        else:
            self._delete('/zones/{0}'.format(identifier))

        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-NSONE-Key': self._get_provider_option('auth_token')
        }
        default_auth = None

        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers,
                                    auth=default_auth)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Vultr"""
from __future__ import absolute_import
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['vultr.com']


def provider_parser(subparser):
    """Configure provider parser for Vultr"""
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for Vultr"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.vultr.com/v1'

    def _authenticate(self):
        payload = self._get('/dns/list')

        if not [item for item in payload if item['domain'] == self.domain]:
            raise Exception('No domain found')

        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            'type': rtype,
            'domain': self.domain_id,
            'name': self._relative_name(name),
            'priority': 0
        }
        if rtype == 'TXT':
            record['data'] = "\"{0}\"".format(content)
        else:
            record['data'] = content
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')
        self._post('/dns/create_record', record)

        LOGGER.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get('/dns/records', {'domain': self.domain_id})
        records = []
        for record in payload:
            processed_record = {
                'type': record['type'],
                'name': "{0}.{1}".format(record['name'], self.domain_id),
                'ttl': record.get('ttl', self._get_lexicon_option('ttl')),
                'content': record['data'],
                'id': record['RECORDID']
            }
            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record['type'] == rtype]
        if name:
            records = [record for record in records if record['name']
                       == self._full_name(name)]
        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {
            'domain': self.domain_id,
            'RECORDID': identifier,
            'ttl': self._get_lexicon_option('ttl')
        }
        # if rtype:
        #     data['type'] = rtype
        if name:
            data['name'] = self._relative_name(name)
        if content:
            if rtype == 'TXT':
                data['data'] = "\"{0}\"".format(content)
            else:
                data['data'] = content

        self._post('/dns/update_record', data)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            data = {
                'domain': self.domain_id,
                'RECORDID': record_id
            }
            self._post('/dns/delete_record', data)

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        default_headers = {
            'Accept': 'application/json',
            # 'Content-Type': 'application/json',
            'API-Key': self._get_provider_option('auth_token')
        }

        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=data,
                                    headers=default_headers)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        if action in ('DELETE', 'PUT', 'POST'):
            # vultr handles succss/failure via HTTP Codes, Only GET returns a response.
            return response.text
        return response.json()



"""Integration tests for EasyDNS"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class EasyDnsProviderTests(TestCase, IntegrationTests):
    """TestCase for EasyDNS"""
    provider_name = 'easydns'
    domain = 'easydnstemp.com'

    def _test_parameters_overrides(self):
        return {'api_endpoint': 'http://sandbox.rest.easydns.net'}

    def _filter_headers(self):
        return ['Authorization']

    def _filter_query_parameters(self):
        return ['_key', '_user']

    # TODO: the following skipped suite and fixtures should be enabled
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return

    @pytest.fixture(autouse=True)
    def _skip_suite(self, request):  # pylint: disable=no-self-use
        if request.node.get_closest_marker('ext_suite_1'):
            pytest.skip('Skipping extended suite')



"""Module provider for hostingde (Hosting.de)"""
from __future__ import absolute_import

import json
import logging
import time
import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['hosting.de']

# be aware to provide an auth_token
# LEXICON_HOSTINGDE_AUTH_TOKEN

def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument(
        "--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Hosting"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://secure.hosting.de/api/dns/v1/json'


    def _authenticate(self):

        response = self._get_zone_config()

        LOGGER.debug('authenticate debug: %s', response)
        if response == []:
            raise Exception('Domain {} not found'.format(self.domain))

        self.domain_id = response.get('id', None)

    # Helper
    def _get_zone_config(self):

        data = ({"filter": {
            "field": "ZoneName",
            "value": self.domain
        }})
        response = self._request(action='POST', url='/zoneConfigsFind', data=data)
        if response != []:
            return response[0]

        return []

    # Normal Behaviour List all records. If filters are provided, send to the API if possible,
    # else apply filter locally. Return value should be a list of records.
    # Return an empty list if no records found
    # type, name and content are used to filter records.
    # caused the fact, that provider will enclose TXT values with ""
    # we have filter content afterwords
    def _list_records(self, rtype=None, name=None, content=None):
        data = {}
        subfilter = []  # used by API to filter
        subfilter.append({"field": "zoneConfigId", "value": self.domain_id})
        if rtype:
            subfilter.append({"field": "RecordType", "value": rtype})
        if name:
            subfilter.append({"field": "RecordName", "value": self._full_name(name)})
        if subfilter:
            data.update(
                {"filter":{"subFilterConnective": "AND", "subFilter": subfilter}})

        LOGGER.debug('list_records filter: %s', data)
        raw_records = self._request(action='POST', url='/recordsFind', data=data)

        processed_records = []

        for record in raw_records:
            processed_record = {
                'type': record['type'],
                'name': self._full_name(record['name']),
                'ttl': record['ttl'],
                'id': record['id'],
                'content': record['content']
                }
            if record['priority']:
                processed_record['priority'] = record['priority']
            processed_record = self._clean_TXT_record(processed_record)

            processed_records.append(processed_record)

        if content:
            processed_records = [record for record in processed_records
                                 if record['content'].lower() == content.lower()]

        LOGGER.debug('list_records: %s', processed_records)
        return processed_records

    # Normal Behavior Create a new DNS record. Return a boolean True if successful.
    # If Record Already Exists Do nothing. DO NOT throw exception.
    # TTL If not specified or set to 0, use reasonable default.
    def _create_record(self, rtype, name, content):
        records = self._list_records(rtype, name, content)
        if records:
            LOGGER.debug('not creating duplicate record: %s', records[0])
            return True

        zone_config = self._get_zone_config()
        priority = self._get_lexicon_option('priority')
        ttl = self._get_lexicon_option('ttl')

        record = {"name": self._full_name(name),
                  "type": rtype,
                  "content": content
                 }
        if ttl:
            if int(ttl) < 60:
                LOGGER.warning("ttl must be minimum 60")
                ttl = 60
            record['ttl'] = int(ttl)
        if priority:
            record['priority'] = int(priority)

        LOGGER.debug('create_record: %s', record)
        data = {"zoneConfig": zone_config, "recordsToAdd": [record]}

        self._request(action='POST', url='/zoneUpdate', data=data)
        return True

    # Normal Behaviour Update a record. Record to be updated can be specified by providing id OR
    # name, type and content. Return a boolean
    # True if successful.
    # TTL:
    #    If not specified, do not modify ttl.
    #    If set to 0, reset to reasonable default.
    # No Match Throw exception?
    # Update a record. use delete & create cause there is no API update call
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            records = [r for r in records if r['id'] == identifier]
        else:
            records = self._list_records(rtype, name, None)

        if not records:
            raise Exception("Record not found")
        if len(records) > 1:
            raise Exception("Record not unique")
        orig_record = records[0]
        orig_id = orig_record['id']

        new_rtype = rtype if rtype else orig_record['type']
        new_name = name if name else orig_record['name']
        new_content = content if content else orig_record['content']

        self._delete_record(orig_id)
        return self._create_record(new_rtype, new_name, new_content)

    # Normal Behaviour Remove a record. Record to be deleted can be
    # specified by providing id OR name, type and content.
    # Return a boolean True if successful.
    # No Match Do nothing. DO NOT throw exception
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record['id'] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_ids)
        records = []
        for record_id in delete_record_ids:
            records.append({'id': record_id})

        zone_config = self._get_zone_config()
        data = {"zoneConfig": zone_config, "recordsToDelete": records}

        self._request(action='POST', url='/zoneUpdate', data=data)

        # sometimes it takes some time to delete a record
        # so, here check after deleting and loop
        # if record(s) is not deleted after 30 seconds > False
        retries = 30
        for record_id in delete_record_ids:
            while True:
                if self._list_records(record_id):
                    if retries < 1:
                        break
                    retries = retries - 1
                    time.sleep(1)
                else:
                    break

        LOGGER.debug('delete_records: %s', retries > 0)
        return retries > 0


    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        data.update({
            "authToken": self._get_provider_option('auth_token')
            })

        read_page = 1
        return_data = []
        retries = 30

        # in some situatons, API uses pagination
        # here we check and if there is pagination, go through all pages
        while True:
            page_data = data
            page_data.update({"page": read_page})

            response = requests.request(action, self.api_endpoint + url,
                                        params=query_params,
                                        data=json.dumps(page_data),
                                        headers={'Content-Type': 'application/json'})
            response.raise_for_status()
            response_json = response.json()  # Work with json Object
            LOGGER.debug('_request response: %s', response_json)
            status = response_json.get('status', '*** no status available ***')
            # if API still busy just wait and retry
            if status == 'error' and\
                response_json.get('errors')[0].get('value', '') == 'blocked':
                if retries < 1:
                    raise Exception('Api error: {0}'.format(response_json.get('errors')))
                retries = retries - 1
                time.sleep(1)
                continue

            if status not in ('success', 'pending'):
                raise Exception('Api error: {0}'.format(response_json.get('errors')))
            # check if there a data object
            read_data = response_json.get('response', {}).get('data', None)

            # if no data object, check if there a records object
            if read_data is None:
                read_data = response_json.get('response', {}).get('records', None)

            # add response data to return data
            if read_data:
                return_data += read_data

            # is there pagination and more data available?
            total_pages = response_json.get('response', {}).get('totalPages', 0)
            if total_pages in (read_page, 0):
                break  # no more data

            read_page = read_page + 1

        return return_data
 


"""Integration tests for OVH"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class OvhProviderTests(TestCase, IntegrationTests):
    """TestCase for OVH"""
    provider_name = 'ovh'
    domain = 'pacalis.net'

    def _filter_headers(self):
        return ['X-Ovh-Application', 'X-Ovh-Consumer', 'X-Ovh-Signature']

    def _test_parameters_overrides(self):
        return {'auth_entrypoint': 'ovh-eu'}



"""Integration tests for HostingDE provider"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests

# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class FooProviderTests(TestCase, IntegrationTests):
    """Integration tests for hostingde provider"""

    provider_name = 'hostingde'
    domain = 'eruza.de'

    def _filter_post_data_parameters(self):
        return ['authToken']

    def _filter_headers(self):
        return ['Authorization']

    def _filter_query_parameters(self):
        return ['secret_key']
